"""Tooling to fit and sample splines."""

import warnings
from typing import Any, Dict, List, Optional, Tuple, Union

import numpy as np
from psygnal import EventedModel
from pydantic import PrivateAttr, conint, root_validator, validator
from scipy.interpolate import splev, splprep
from scipy.spatial.transform import Rotation, Slerp

from .utils import coaxial_y_vectors_from_z_vectors, within_range, get_mask_limits


class NDimensionalSpline(EventedModel):
    """Model for multidimensional splines."""

    points: np.ndarray
    order: conint(ge=1, le=5) = 3
    smoothing: Optional[int] = None
    mask: Optional[np.ndarray] = None
    closed: bool = False
    _n_spline_samples: int = PrivateAttr(10000)
    _tck = PrivateAttr(Tuple)
    _u_mask_limits: List[Tuple[float, float]] = PrivateAttr([])
    _length = PrivateAttr(float)

    class Config:
        """Pydantic BaseModel configuration."""

        arbitrary_types_allowed = True

    def __init__(self, **kwargs):
        """Calculate the splines after validating the paramters."""
        super().__init__(**kwargs)
        self._prepare_spline()

    @property
    def _ndim(self) -> int:
        ndim: int = self.points.shape[1]
        return ndim

    @validator("points", pre=True)
    def is_coordinate_array(cls, v: Union[List[List[float]], np.ndarray]) -> np.ndarray:
        """Validate and coerce the points values to a 2D numpy array."""
        points = np.atleast_2d(v)
        if points.ndim != 2:
            raise ValueError("points must be an (n, d) array")
        if len(points) < 2:
            raise ValueError("must provide at least 2 points")
        return points

    @root_validator(skip_on_failure=True)
    def validate_number_of_points(
        cls, values: Dict[str, Union[np.ndarray, int]]
    ) -> Dict[str, Union[np.ndarray, int]]:
        """Verify that the number of points > spline_order."""
        points: np.ndarray = values.get("points")
        n_samples = points.shape[0]
        spline_order: Optional[int] = values.get("order")

        if spline_order is not None and n_samples <= spline_order:
            new_order = n_samples - 1
            warnings.warn(
                f'Too few points for spline of order {spline_order}. '
                f'Decreasing order to {new_order}'
            )
            values['order'] = new_order

        return values

    def __setattr__(self, name: str, value: Any) -> None:
        """Overwritten so that splines are recalculated when points are updated."""
        super().__setattr__(name, value)
        if name in ("points", "order", "smoothing", "mask"):  # ensure splines stay in sync
            self._prepare_spline()

    def _prepare_spline(self) -> None:
        """Spline parametrisation mapping [0, 1] to a smooth curve through spline points.

        Equidistant samples between 0 and 1 will yield points equidistant along
        the spline in euclidean space.
        """
        # oversample an initial spline to ensure better distance parametrisation
        u = np.linspace(0, 1, self._n_spline_samples)

        if self.closed:
            points = np.append(self.points, self.points[:1], axis=0)
        else:
            points = self.points

        tck, raw_u = splprep(points.T, s=0, k=self.order, per=self.closed)
        samples = np.stack(splev(u, tck), axis=1)

        # calculate the cumulative length of line segments
        # as we move along the filament.
        inter_point_differences = np.diff(samples, axis=0)
        inter_point_distances = np.linalg.norm(inter_point_differences, axis=1)
        cumulative_distance = np.cumsum(inter_point_distances)
        # prepend a zero, no distance has been covered
        # at start of spline parametrisation
        cumulative_distance = np.insert(cumulative_distance, 0, 0)
        # save length for later and normalize
        self._length = cumulative_distance[-1]

        if self.mask is not None:
            limits = get_mask_limits(self.mask)
            u_ranges = []
            for low_idx, high_idx in limits:
                u_low, u_high = raw_u[[low_idx, high_idx]]
                u_ranges.append((u_low, u_high))
            self._u_mask_limits = u_ranges

        cumulative_distance /= self._length

        self._tck, u = splprep(
            samples.T, u=cumulative_distance, s=0, k=self.order
        )

    def sample(
        self,
        u: Optional[Union[float, np.ndarray]] = None,
        separation: Optional[float] = None,
        n_samples: Optional[int] = None,
        derivative_order: int = 0,
    ) -> np.ndarray:
        """Sample points or derivatives on the spline.

        This function yields samples equidistant in Euclidean space
        along the spline for linearly spaced values of u.
        Only one of u, separation or n_samples should be provided.

        Parameters
        ----------
        u : Optional[Union[float, np.ndarray]]
            The positions to sample the spline at. These are in the normalized
            spline coordinate, which spans [0, 1].
        separation : Optional[float]
            The desired separation between sampling points in Euclidean space.
        n_samples : Optional[int]
            The total number of equidistant points to sample along the spline.
        derivative_order : int
            Order of the derivative to evaluate at each spline position.
            If 0, the position on the spline is returned.
            If greater than 0, the derivative of position is returned
            (e.g., 1 for tangent vector). The derivative_order must be less than
            or equal to the spline order. The default value is 0.

        Returns
        -------
        values : np.ndarray
            The values along the spline.
            If derivative_order is 0, returns positions.
            If calculate_derivative >0, returns derivative vectors.
        """
        if (derivative_order < 0) or (derivative_order > self.order):
            raise ValueError("derivative order must be [0, spline_order]")
        if sum(arg is not None for arg in (u, separation, n_samples)) != 1:
            raise ValueError("only one of u, separation or n_samples should be provided.")
        if u is None:
            u = self._get_equidistant_spline_coordinate_values(separation=separation, n_samples=n_samples)

        samples = splev(np.atleast_1d(u), self._tck, der=derivative_order)
        samples = np.stack(samples, axis=1)  # (n, d)

        return samples

    def _get_equidistant_spline_coordinate_values(
        self,
        separation: Optional[float] = None,
        n_samples: Optional[int] = None,
        approximate: bool = False,
    ) -> np.ndarray:
        """Calculate spline coordinates for points with a given Euclidean separation.

        Only one of separation or n_samples should be provided.

        Parameters
        ----------
        separation : float
            The Euclidean distance between desired spline samples.
        n_samples : int
            The total number of equidistant points to sample along the spline.
        approximate : bool
            Approximate the separation in order to include the extrema.
            Has no effect if n_samples is provided.

        Returns
        -------
        u : np.ndarray
            The array of spline coordinate values.
        """
        if separation is not None and n_samples is not None:
            raise ValueError("only one of separation and n_samples should be provided.")

        if n_samples is not None:
            remainder = 0
        elif separation is not None:
            n_samples = int(self._length / separation)
            if n_samples == 0:
                raise ValueError(f'separation ({separation}) must be less than '
                                 f'length ({self._length})')
            if approximate:
                remainder = 0
            else:
                remainder = (self._length % separation) / self._length

        return np.linspace(0, 1 - remainder, n_samples)

    def get_mask_for_u(self, u):
        """Get a 1D boolean mask for a given u.

        The mask is True where values should be kept.
        """
        # we need to be careful here because there are precision issues with splines
        # and sometimes we get values below or above the limits when they should be equal
        # so we use a special within_range function
        u = np.atleast_1d(u)
        mask = np.zeros_like(u, bool)
        for low, high in self._u_mask_limits:
            mask[within_range(u, low, high)] = 1
        return mask

    def reverse(self):
        """Reverse the order of points and recompute the spline."""
        self.points = self.points[::-1, :]


class Spline3D(NDimensionalSpline):
    """3D spline model with interpolation of orientations along the spline.

    Basis vectors for a local coordinate system along the spline can be calculated.
    In this coordinate system, z vectors are tangent to the spline and the xy-plane
    changes minimally and smoothly.
    """

    _rotation_sampler = PrivateAttr(Slerp)

    @validator("points")
    def _is_3d_coordinate_array(cls, v):
        if v.ndim != 2 or v.shape[-1] != 3:
            raise ValueError("must be an (n, 3) array")
        return v

    def _prepare_spline(self):
        super()._prepare_spline()
        self._prepare_orientation_sampler()

    def _prepare_orientation_sampler(self):
        """Prepare a pose_sampler yielding smoothly varying orientations along the spline.

        This method constructs a set of rotation matrices which vary smoothly with
        the spline coordinate `u`. A pose_sampler is then prepared which can be queried at
        any point(s) along the spline coordinate `u` and the resulting rotations vary
        smoothly along the spline
        """
        u = np.linspace(0, 1, num=self._n_spline_samples)
        z = self._sample_spline_z(u)
        y = coaxial_y_vectors_from_z_vectors(z)
        x = np.cross(y, z)
        r = Rotation.from_matrix(np.stack([x, y, z], axis=-1))
        self._rotation_sampler = Slerp(u, r)

    def sample_orientations(
        self,
        u: Optional[Union[float, np.ndarray]] = None,
        separation: Optional[float] = None,
        n_samples: Optional[int] = None,
    ) -> Rotation:
        """Local coordinate system at any point along the spline.


        Only one of u, separation or n_samples should be provided.
        """
        if sum(arg is not None for arg in (u, separation, n_samples)) != 1:
            raise ValueError("only one of u, separation or n_samples should be provided.")
        if u is None:
            u = self._get_equidistant_spline_coordinate_values(separation=separation, n_samples=n_samples)

        rot = self._rotation_sampler(u)
        if rot.single:
            rot = Rotation.concatenate([rot])

        return rot

    def _sample_spline_z(self, u: Union[float, np.ndarray]) -> np.ndarray:
        """Sample vectors tangent to the spline."""
        z = self.sample(u, derivative_order=1)
        z /= np.linalg.norm(z, axis=1, keepdims=True)
        return z

    def _sample_spline_y(self, u: Union[float, np.ndarray]) -> np.ndarray:
        """Sample vectors perpendicular to the spline."""
        rotations = self.sample_orientations(u)
        return rotations.as_matrix()[..., 1]



import numpy as np
from scipy.spatial import KDTree

from morphosamplers import MorphoModels
from morphosamplers.samplers import sphere_samplers


def test_point_sampler():
    """Should produce points with correct spacing on surface of sphere."""
    sphere = MorphoModels.Sphere(center=(5, 5, 5), radius=10)
    sampler = sphere_samplers.PointSampler(spacing=2)
    points = sampler.sample(sphere)

    # all points should be centered around (5, 5, 5)
    eps = 1e-2
    assert np.mean(points) - 5 < eps

    # spacing between points should be ~2
    tree = KDTree(points)
    distances, idx = tree.query(points, k=4)
    distances = distances[:, 1:]  # exclude distance to self
    assert 1.75 <= np.mean(distances) < 2.25


def test_pose_sampler():
    """Should produce poses oriented with Z normal to the surface of the sphere."""
    sphere = MorphoModels.Sphere(center=(5, 5, 5), radius=10)
    sampler = sphere_samplers.PoseSampler(spacing=2)
    poses = sampler.sample(sphere)

    # all points should be centered around (5, 5, 5)
    eps = 1e-2
    assert np.mean(poses.positions) - 5 < eps

    # points should be oriented with z normal to the sphere
    z_vecs_poses = poses.orientations[:, :, 2]
    z_dir_expected = poses.positions - np.array([5, 5, 5])
    z_dir_expected /= np.linalg.norm(z_dir_expected, axis=-1, keepdims=True)
    assert np.allclose(z_vecs_poses, z_dir_expected)

    # check that rotation matrices are in SO3
    assert np.allclose(np.linalg.det(poses.orientations), 1)



import einops
import numpy as np

from morphosamplers.sample_types import PoseSet
from morphosamplers.samplers.path_samplers import PathSamplers
from morphosamplers.models import MorphoModels


def test_instantiation():
    sampler = PathSamplers.PoseSampler(spacing=5)
    assert isinstance(sampler, PathSamplers.PoseSampler)


def test_sampling():
    """Samples should interpolate control points with consistent orientations."""
    n_points = 50
    total_length = 100
    spacing = total_length / n_points
    x = y = np.zeros(shape=(n_points, ))
    z = np.linspace(0, total_length, num=n_points)
    control_points = einops.rearrange([x, y, z], 'xyz b -> b xyz')
    path = MorphoModels.Path(control_points=control_points)
    sampler = PathSamplers.PoseSampler(spacing=spacing)
    samples = sampler.sample(path)

    # positions should be an interpolation along z from 0 -> total_length
    assert isinstance(samples, PoseSet)
    expected = np.linspace([0, 0, 0], [0, 0, total_length], num=n_points)
    assert np.allclose(samples.positions, expected)

    # z axis should be a unit vector along +z
    z = samples.orientations[:, :, 2]
    assert np.allclose(z, [0, 0, 1])

    # y0 and y1 should be ~identical unit vectors...
    y0 = samples.orientations[0, :, 1]
    y1 = samples.orientations[1, :, 1]
    projection = np.dot(y0, y1)
    assert np.allclose(np.linalg.norm(y0), 1)
    assert np.allclose(np.linalg.norm(y1), 1)
    assert np.allclose(projection, 1)

    # check for determinant 1 on all rotation matrices
    determinants = np.linalg.det(samples.orientations)
    assert np.allclose(determinants, 1)



from typing import List

import numpy as np
from pydantic import validator

from morphosamplers.core import MorphoModel


class Surface(MorphoModel):
    """A 3D surface defined by control points in a series of levels."""
    control_points: List[np.ndarray]

    @validator('control_points')
    def check_at_least_two_points(cls, value):
        if len(value) < 2:
            raise ValueError('A Path must contain at least two levels.')
        for array in value:
            if len(array) < 2:
                raise ValueError('Each level must contain at least two points.')
        return value

    @validator('control_points', pre=True)
    def ensure_list_of_float_arrays(cls, value):
        return [np.asarray(v, dtype=np.float32) for v in value]

    def __len__(self) -> int:
        return len(self.control_points)


import numpy as np
from scipy.spatial.transform import Rotation as R

from morphosamplers.core import MorphoSampler
from morphosamplers.models import Path
from morphosamplers.sample_types import PoseSet

from .pose_sampler_parallel import PoseSampler


class HelicalPoseSampler(MorphoSampler):
    """Sample poses along the backbone of a helical path.

    In the sample poses
    - positions will be separated by `spacing`.
    - z-axis at each position will be aligned with the filament axis
    - xy-planes will be rotated by `twist` degrees between adjacent positions.
    """
    spacing: float  # spacing between positions
    twist: float  # helical twist per subunit in degrees

    def sample(self, obj: Path) -> PoseSet:
        # get parallel poses along filament axis
        sampler = PoseSampler(spacing=self.spacing)
        poses = sampler.sample(obj)

        # rotate poses so they respect helical parameters
        angles = np.linspace(start=0, stop=self.twist * len(poses), num=len(poses))
        Rz = R.from_euler('z', angles=angles, degrees=True).as_matrix()
        new_orientations = poses.orientations @ Rz
        return PoseSet(positions=poses.positions, orientations=new_orientations)



import numpy as np
from pydantic import BaseModel, root_validator, validator
from typing import Union
from typing_extensions import TypeAlias

Image: TypeAlias = np.ndarray
Mask: TypeAlias = np.ndarray
Points: TypeAlias = np.ndarray  # (n, 3)


class PoseSet(BaseModel):
    """Model for a set of 3D poses."""
    positions: np.ndarray  # (n, 3)
    orientations: np.ndarray  # (n, 3, 3)

    class Config:
        allow_mutation = False
        arbitrary_types_allowed = True

    @validator('positions', pre=True)
    def coerce_to_n_by_3_array(cls, value):
        value = np.atleast_2d(np.asarray(value))
        if value.ndim != 2 or value.shape[-1] != 3:
            raise ValueError('positions must be an (n, 3) array.')
        return value

    @validator('orientations', pre=True)
    def check_n_by_3_3_array(cls, value):
        value = np.asarray(value)
        if value.ndim != 3 or value.shape[-2:] != (3, 3):
            raise ValueError('orientations must be an (n, 3, 3) array.')
        return value

    @root_validator
    def check_same_length(cls, values):
        positions, orientations = values.get('positions'), values.get('orientations')
        if len(positions) != len(orientations):
            raise ValueError("lengths of positions and orientations don't match")
        return values

    def __len__(self) -> int:
        return len(self.positions)


SampleType = Union[PoseSet, Image, Mask, Points]



import einops
import numpy as np
from scipy.interpolate import splev
from scipy.spatial.transform import Rotation, Slerp

from morphosamplers import Path
from morphosamplers.core import MorphoSampler
from morphosamplers.sample_types import PoseSet
from morphosamplers.utils import coaxial_y_vectors_from_z_vectors


class PoseSampler(MorphoSampler):
    spacing: float

    def sample(self, obj: Path) -> PoseSet:
        """Sample a `Path` to produces a `PoseSet`.

        The positions in the `PoseSet` are equidistant points sampled according to
        `spacing`. The orientations in the `PoseSet` are as close to parallel to each
        other as curvature of the path allows. The z-axis is oriented along a
        smooth curve through the control points.
        """
        # sample equidistant points along path
        from .point_sampler import PointSampler
        tck, total_length = PointSampler.prepare_spline(obj)
        n_samples = total_length // self.spacing
        max_u = 1 - ((total_length % self.spacing) / total_length)
        u = np.linspace(0, max_u, num=int(n_samples))
        positions = einops.rearrange(splev(u, tck), 'xyz b -> b xyz')

        # oversample orientations along path which are maximally coaxial (parallel)
        u_oversampled = np.linspace(0, 1, num=5000)
        z = einops.rearrange(splev(u_oversampled, tck, der=1), 'xyz b -> b xyz')
        z /= np.linalg.norm(z, axis=-1, keepdims=True)
        y = coaxial_y_vectors_from_z_vectors(z)
        x = np.cross(y, z)
        x /= np.linalg.norm(x, axis=-1, keepdims=True)
        rotations = np.empty(shape=(5000, 3, 3))
        rotations[:, :, 0] = x
        rotations[:, :, 1] = y
        rotations[:, :, 2] = z
        # rotations = einops.rearrange([x, y, z], 'xyz b vec -> b vec xyz')

        # interpolate rotations
        rotation_sampler = Slerp(u_oversampled, Rotation.from_matrix(rotations))
        orientations = rotation_sampler(u).as_matrix()
        return PoseSet(positions=positions, orientations=orientations)



import numpy as np
from pydantic import validator

from morphosamplers.core import MorphoModel


class Path(MorphoModel):
    """A 3D path defined by an `(n, 3)` array of control points."""
    control_points: np.ndarray

    @validator('control_points', pre=True)
    def coerce_to_n_by_3_array(cls, value):
        value = np.atleast_2d(np.asarray(value))
        if value.ndim != 2 or value.shape[-1] != 3:
            raise ValueError('`control_points` must be an (n, 3) array.')
        return value

    @validator('control_points')
    def check_at_least_two_points(cls, value):
        if len(value) < 2:
            raise ValueError('A Path must contain at least two points.')
        return value

    def __len__(self) -> int:
        return len(self.control_points)



from scipy.spatial.transform import Rotation
import numpy as np

from .spline import Spline3D


class HelicalFilament(Spline3D):

    def sample_helical(self, rise, twist=0, radial_offset=0, cyclic_symmetry_order=1, twist_offset=0, degrees=True):
        positions = []
        orientations = []
        base_positions = self.sample(separation=rise)
        base_orientations = self.sample_orientations(separation=rise)

        # twist around the z acis by the given angle
        twist_angles = twist_offset + (np.arange(len(base_positions)) * twist)
        Rz = Rotation.from_euler('z', angles=twist_angles, degrees=degrees)
        base_orientations *= Rz

        # create symmetry duplicates around the axis of the filament
        for i in range(cyclic_symmetry_order):
            # rotate around z by i * 360/c-symmetry-group
            twist_angles = i * (2 * np.pi / cyclic_symmetry_order)
            Rz = Rotation.from_euler('z', angles=twist_angles)
            ori = base_orientations * Rz
            # shift all particles away from the filament axis by self.radius
            # direction of shift is determined by the twist of the particle (along its y axis)
            y = ori.as_matrix()[..., 1]
            pos = base_positions + y * radial_offset
            positions.append(pos)
            orientations.append(ori)

        return np.concatenate(positions, axis=0), np.concatenate(orientations, axis=0)

    @staticmethod
    def subunits_per_turn(self, twist):
        """Number of subunits per full helical turn."""
        if twist == 0:
            return 1.0
        return 360 / np.abs(twist)

    @staticmethod
    def pitch(self, twist, rise):
        """Helical pitch, the vertical distance along the helix for one full turn."""
        if twist == 0:
            return rise
        return 360 * rise / np.abs(twist)

    @staticmethod
    def handedness(self, twist):
        if twist == 0:
            return 1
        return np.sign(twist)



import einops
import numpy as np

from morphosamplers import PathSamplers
from morphosamplers import MorphoModels


def test_instantiation():
    sampler = PathSamplers.PointSampler(spacing=5)
    assert isinstance(sampler, PathSamplers.PointSampler)


def test_sampling():
    """Samples should interpolate control points"""
    n_points = 50
    total_length = 100
    spacing = total_length / n_points
    x = y = np.zeros(shape=(n_points, ))
    z = np.linspace(0, total_length, num=n_points)
    control_points = einops.rearrange([x, y, z], 'xyz b -> b xyz')
    path = MorphoModels.Path(control_points=control_points)
    sampler = PathSamplers.PointSampler(spacing=spacing)
    samples = sampler.sample(path)
    assert isinstance(samples, np.ndarray)
    assert samples.ndim == 2
    expected = np.linspace([0, 0, 0], [0, 0, total_length], num=n_points)
    assert np.allclose(samples, expected)


"""Tooling to fit and sample surfaces."""

import warnings
from typing import List, Optional, Tuple, Union

import einops
import numpy as np
from psygnal import EventedModel
from pydantic import PrivateAttr, conint, root_validator, validator
from scipy.interpolate import splev, splprep
from scipy.spatial.transform import Rotation

from .spline import Spline3D
from .utils import (
    estimate_point_row_directions,
    extrapolate_point_rows_with_derivatives,
    minimize_closed_point_row_pair_distance,
    minimize_point_row_pair_distance,
)


class _SplineSurface(EventedModel):
    """Surface model based on splines."""

    points: List[np.ndarray]
    separation: float
    order: conint(ge=1, le=5) = 3
    smoothing: Optional[int] = None
    closed: bool = False
    inside_point: Optional[Union[np.ndarray, Tuple[float, float, float]]] = None
    oversampling: int = 2
    _row_splines = PrivateAttr(List[Spline3D])
    _column_splines = PrivateAttr(List[Spline3D])

    class Config:
        """Pydantic BaseModel configuration."""

        arbitrary_types_allowed = True

    @validator("points")
    def _validate_number_of_splines(v):
        points = np.atleast_2d(*v)
        if len(points) < 2:
            raise ValueError(
                "At least 2 arrays of points are necessary to define a surface."
            )
        return points

    @root_validator(skip_on_failure=True)
    def _validate_number_of_lines(cls, values):
        points = values.get("points")
        n_lines = len(points)
        order = values.get("order")

        # order needs to be reduced if it does not match the number of splines
        if order is not None and n_lines <= order:
            new_order = n_lines - 1
            warnings.warn(
                f"Too few arrays of points for interpolation of order {order}. "
                f"Decreasing order to {new_order} for interpolation between lines."
            )
            # we don't decrease the order here so we can still attempt to use the full
            # interpolation for individual splines, so we do it in _generate_column_splines

        return values

    def __init__(self, **kwargs):
        """Calculate the splines after validating the paramters."""
        super().__init__(**kwargs)
        self._prepare_splines()

    def __setattr__(self, name, value):
        super().__setattr__(name, value)
        if name in ("points", "separation", "order", "closed", "oversampling"):
            self._prepare_splines()

    def _prepare_splines(self):
        self._generate_row_splines()
        self._generate_column_splines()

    @classmethod
    def _fix_spline_edges(cls, splines, separation, order, closed, oversampling):
        """
        Generate new control points for each spline to ensure a minimally deformed grid.

        For open surfaces, splines are sampled with the same equidistance, and the resulting
        rows of points are then padded with NaNs at each end so that:
        - they have the same amount of samples
        - mean distance between samples at the same index in neighboring rows is minimised
        NaN values are then replaced by extrapolating the existing values along the spline
        derivative.

        For closed surfaces, since they cannot be extended, each spline is sampled with the same
        number of points, which may result in a deformed grid. Each row is then aligned to minimise
        shear by minimising the mean distance between samples at the same intex in neighboring rows.
        """
        # sample splines to get equidistant points on z slices
        separation = separation / oversampling
        us = [
            spline._get_equidistant_spline_coordinate_values(
                separation=separation, approximate=True
            )
            for spline in splines
        ]

        if closed:
            # We need the same amount of samples on each row to make a rectangular grid,
            # but we can't extend/pad a closed surface, so instead we take n_samples from each
            # so that on average we get a good approximation of the given separation
            n_points = [len(u) for u in us]
            diff = np.max(n_points) - np.min(n_points)
            if diff > 1:
                warnings.warn(
                    "The grid is deformed by more than 1 separation in some places. "
                    "This is inevitable with closed, non-cylindrical grids."
                )
            best_n = round(np.mean(n_points))

            points = [spline.sample(n_samples=best_n) for spline, u in zip(splines, us)]

            points = minimize_closed_point_row_pair_distance(
                points, expected_dist=separation
            )
            masks = [np.ones(len(p), dtype=bool) for p in points]
        else:
            points = [spline.sample(u) for spline, u in zip(splines, us)]
            derivatives = [
                spline.sample(u, derivative_order=1) for spline, u in zip(splines, us)
            ]

            # flip directions of annotations if necessary
            directions = estimate_point_row_directions(points)
            # coordinates can be simply inverted
            points = [pts[::dir] for pts, dir in zip(points, directions)]
            # directions need to be inverted *and* flipped (since they are a vector)
            derivatives = [
                der[::dir] * dir for der, dir in zip(derivatives, directions)
            ]

            # extrapolate where nans are present by extending along the spline direction
            points = minimize_point_row_pair_distance(points, expected_dist=separation)

            masks = [~np.isnan(p[:, 0]) for p in points]  # just one dim is enough
            points = extrapolate_point_rows_with_derivatives(
                points, derivatives, separation
            )

        return points, masks

    def _generate_row_splines(self):
        self._row_splines = [
            Spline3D(
                points=p, order=self.order, smoothing=self.smoothing, closed=self.closed
            )
            for p in self.points
        ]

    def _generate_column_splines(self):
        # fix edge artifacts by extending splines until we have a grid
        control_points, masks = self._fix_spline_edges(
            self._row_splines,
            self.separation,
            self.order,
            self.closed,
            self.oversampling,
        )

        if self.closed:
            # _fix_spline_edges returns strips including the extrema, so we need to remove
            # one of them if the surface is closed to avoid duplication
            control_points = [p[:-1] for p in control_points]

        # _fix_spline_edges oversamples to avoid artifacts, so we only take every N points
        control_points = np.stack(control_points)[:, ::self.oversampling]
        masks = np.stack(masks)[:, ::self.oversampling]

        # stack points in the other direction, so we get the column-splines
        stacked = einops.rearrange(control_points, "row column xyz -> column row xyz")
        masks_stacked = einops.rearrange(masks, "row column -> column row")

        # order needs to be reduced if it does not match the number of splines
        if len(self.points) <= self.order:
            order = len(self.points) - 1
        else:
            order = self.order

        self._column_splines = [
            Spline3D(points=p, order=order, smoothing=self.smoothing, mask=mask)
            for p, mask in zip(stacked, masks_stacked)
        ]


class GriddedSplineSurface(_SplineSurface):
    """Surface model defined by a regular grid of splines."""

    _grid_splines = PrivateAttr(List[Tuple])
    _grid_meta_splines = PrivateAttr(List[Tuple])
    _mask = PrivateAttr(np.ndarray)

    def _prepare_splines(self):
        super()._prepare_splines()
        self._generate_grid_splines()

    @property
    def grid_shape(self):
        return (len(self._grid_splines), len(self._grid_meta_splines))

    def _generate_grid_splines(self):
        # we finally create grid-like splines with optimized spacing
        # though not optimal for consistent spacing, we have to use the same n
        # for each spline, otherwise we end up with offset points in each lines
        separation = self.separation / self.oversampling
        us = [
            spline._get_equidistant_spline_coordinate_values(
                separation=separation, approximate=True
            )
            for spline in self._column_splines
        ]
        n_points = [len(u) for u in us]
        diff = np.max(n_points) - np.min(n_points)
        if diff > 1:
            warnings.warn(
                "The grid is deformed by more than 1 separation in some places. "
                "Consider passing points that belong to parallel planes, or splitting "
                "the surface in quasi-planar patches."
            )
        best_n = int(round(np.mean(n_points)))
        u = np.linspace(0, 1, best_n)
        # TODO: actually use non-approximate linspace (with remainder) so we get as close as possible to
        #       a real grid. The problem with this is that we won't reach exactly the last spline

        equidistant_points = [spline.sample(u)[::self.oversampling] for spline in self._column_splines]
        masks = [spline.get_mask_for_u(u)[::self.oversampling] for spline in self._column_splines]

        self._mask = np.concatenate(np.stack(masks, axis=1))

        # these last splines should not be oversampled, because we want exact
        # positions for our knots, which we save in self._meta_meta_splines_u
        splines = []
        for p in np.stack(equidistant_points, axis=1):
            splines.append(splprep(p.T, s=0, k=self.order))
        self._grid_splines = splines

        # then generate the column grid-splines to ensure they point exactly to the same
        # coordinates (the knots are shared), so we can get perfect orientations
        splines = []
        for p in np.stack(equidistant_points, axis=0):
            splines.append(splprep(p.T, s=0, k=self.order))
        self._grid_meta_splines = splines

    def sample(self) -> np.ndarray:
        """Sample an approximately equidistant grid of points on the surface.

        Samples are optimized for consistent separation and grid-like ordering, which
        results in many discarded edges if the input differs a lot from a rectangle.

        The sampled array has shape (rows * columns, 3) for a grid of shape
        (rows, columns) and follows the (rows, columns) order.
        """
        equidistant_points = [
            einops.rearrange(splev(u, tck), "xyz column -> column xyz")
            for tck, u in self._grid_splines
        ]
        return np.concatenate(equidistant_points)

    @property
    def mask(self):
        return self._mask

    def sample_orientations(self):
        """Sample an approximately equidistant grid of orientations on the surface.

        Follows the same pattern as GriddedSplineSurface.sample(). Orientations are
        generated so that the basis z vector is aligned to the surface normal, while
        x and y vectors are aligned to the row and column splines respectively.
        """
        equidistant_x_vecs = [
            einops.rearrange(splev(u, tck, der=1), "xyz column -> column xyz")
            for tck, u in self._grid_splines
        ]
        equidistant_y_vecs = [
            einops.rearrange(splev(u, tck, der=1), "xyz row -> row xyz")
            for tck, u in self._grid_meta_splines
        ]

        x = einops.rearrange(equidistant_x_vecs, "row column xyz -> (row column) xyz")
        y = einops.rearrange(equidistant_y_vecs, "column row xyz -> (row column) xyz")

        y /= np.linalg.norm(y, axis=1, keepdims=True)
        x /= np.linalg.norm(x, axis=1, keepdims=True)

        z = np.cross(x, y)
        z /= np.linalg.norm(z, axis=1, keepdims=True)

        # since x and y are not guaranteed to be orthogonal, we re-generate y to get
        # perfectly orthogonal so we end up with determinant == 1
        x = np.cross(y, z)
        x /= np.linalg.norm(x, axis=1, keepdims=True)

        mat = np.stack([x, y, z], axis=-1)
        rots = Rotation.from_matrix(mat)

        if self.inside_point is not None:
            # use the inside point to put normal in correct direction
            pts = self.sample()
            # get closest neighbour
            diffs = pts - self.inside_point
            closest_idx = np.argmin(np.linalg.norm(diffs, axis=1))
            z_closest = z[closest_idx]
            z_inside = diffs[closest_idx]
            # flip is the point is outside
            if np.dot(z_inside, z_closest) < 0:
                rots *= Rotation.from_euler("x", np.pi)

        return rots

    def mesh(self) -> Tuple[np.array, np.array]:
        """
        A mesh representation of the surface.

        For each quad in the grid, indices are generated following this pattern:
        0--2  x--2
        | /|  | /|
        |/ |  |/ |
        1--x  0--1

        Returns
        -------
        Tuple[np.array, np.array]
            Vertices coordinates (n, 3) and indices of vertices forming triangle faces (m, 3)
        """
        points = self.sample()
        rows, columns = self.grid_shape
        row_range = np.arange(rows)
        column_range = np.arange(columns)
        shift = 0
        if self.closed:
            columns += 1
            shift = 1
            column_range = np.append(column_range, 0)

        # first half of triangles
        first_index = np.repeat(row_range[:-1], columns - 1) * (columns - shift) + np.tile(column_range[:-1], rows - 1)
        second_index = np.repeat(row_range[1:], columns - 1) * (columns - shift) + np.tile(column_range[:-1], rows - 1)
        third_index = np.repeat(row_range[:-1], columns - 1) * (columns - shift) + np.tile(column_range[1:], rows - 1)
        triangles_1 = np.stack([first_index, second_index, third_index], axis=1)

        # second half
        first_index = np.repeat(row_range[1:], columns - 1) * (columns - shift) + np.tile(column_range[:-1], rows - 1)
        second_index = np.repeat(row_range[1:], columns - 1) * (columns - shift) + np.tile(column_range[1:], rows - 1)
        third_index = np.repeat(row_range[:-1], columns - 1) * (columns - shift) + np.tile(column_range[1:], rows - 1)
        triangles_2 = np.stack([first_index, second_index, third_index], axis=1)

        all_triangles = np.concatenate([triangles_1, triangles_2])
        return points, all_triangles


class HexSplineSurface(_SplineSurface):
    """Surface model defined by a grid of splines which generate a hex grid of samples."""

    _row_splines_even = PrivateAttr(List[Tuple])
    _row_splines_odd = PrivateAttr(List[Tuple])
    _column_splines_even = PrivateAttr(List[Tuple])
    _column_splines_odd = PrivateAttr(List[Tuple])
    _mask = PrivateAttr(np.ndarray)

    def _prepare_splines(self):
        super()._prepare_splines()
        self._generate_offset_splines()

    def _generate_offset_splines(self):
        # we finally create grid-like splines with optimized spacing
        # though not optimal for consistent spacing, we have to use the same n
        # for each spline, otherwise we end up with offset points in each lines
        separation = self.separation / self.oversampling
        us = [
            spline._get_equidistant_spline_coordinate_values(
                separation=separation, approximate=True
            )
            for spline in self._column_splines
        ]
        best_n = int(np.mean([len(u) for u in us]))

        us = []
        for i, spline in enumerate(self._column_splines):
            if i % 2:
                us.append(np.linspace(0, 1, best_n))
            else:
                offset = self.separation / spline._length / 2
                us.append(np.linspace(offset, 1 - offset, best_n - 1))

        equidistant_points = [
            spline.sample(u)[::self.oversampling] for spline, u in zip(self._column_splines, us)
        ]

        masks = [
            spline.get_mask_for_u(u)[::self.oversampling] for spline, u in zip(self._column_splines, us)
        ]
        masks_even = np.concatenate(np.stack(masks[::2], axis=1))
        masks_odd = np.concatenate(np.stack(masks[1::2], axis=1))
        self._mask = np.concatenate([masks_even, masks_odd])

        # alternate each row spline to have points on same level
        splines_even = []
        for p in np.stack(equidistant_points[::2], axis=1):
            splines_even.append(splprep(p.T, s=0, k=self.order))
        self._row_splines_even = splines_even

        splines_odd = []
        for p in np.stack(equidistant_points[1::2], axis=1):
            splines_odd.append(splprep(p.T, s=0, k=self.order))
        self._row_splines_odd = splines_odd

        # then alternate columns splines
        splines_even = []
        for p in np.stack(equidistant_points[::2], axis=0):
            splines_even.append(splprep(p.T, s=0, k=self.order))
        self._column_splines_even = splines_even
        splines_odd = []
        for p in np.stack(equidistant_points[1::2], axis=0):
            splines_odd.append(splprep(p.T, s=0, k=self.order))
        self._column_splines_odd = splines_odd

    def sample(self):
        """Sample an approximately equidistant grid of points on the surface.

        Samples are optimized for consistent separation and grid-like ordering, which
        results in many discarded edges if the input differs a lot from a rectangle.
        """
        pts = []
        for splines in (self._row_splines_even, self._row_splines_odd):
            equidistant_points = [np.stack(splev(u, tck), axis=1) for tck, u in splines]
            pts.append(np.concatenate(equidistant_points))
        return np.concatenate(pts)

    @property
    def mask(self):
        return self._mask

    def sample_orientations(self):
        """Sample an approximately equidistant grid of orientations on the surface."""
        rots = []
        for splines, meta_splines in (
            (self._row_splines_even, self._column_splines_even),
            (self._row_splines_odd, self._column_splines_odd),
        ):
            equidistant_x_vecs = [
                np.stack(splev(u, tck, der=1), axis=1) for tck, u in splines
            ]
            equidistant_y_vecs = [
                np.stack(splev(u, tck, der=1), axis=1) for tck, u in meta_splines
            ]

            x = np.concatenate(equidistant_x_vecs)
            # y vectors are generated on column-by-column basis, so we need to swap axes
            y = einops.rearrange(equidistant_y_vecs, "column row xyz-> row column xyz")
            y = np.concatenate(equidistant_y_vecs)

            x /= np.linalg.norm(x, axis=1, keepdims=True)
            y /= np.linalg.norm(y, axis=1, keepdims=True)

            z = np.cross(x, y)
            z /= np.linalg.norm(z, axis=1, keepdims=True)

            rots.append(Rotation.from_matrix(np.stack([x, y, z], axis=-1)))

        rots = Rotation.concatenate(rots)

        if self.inside_point is not None:
            # use the inside point to put normal in correct direction
            pts = self.sample()
            # get closest neighbour
            diffs = pts - self.inside_point
            closest_idx = np.argmin(np.linalg.norm(diffs, axis=1))
            z_closest = z[closest_idx]
            z_inside = diffs[closest_idx]
            # flip is the point is outside
            if np.dot(z_inside, z_closest) < 0:
                rots *= Rotation.from_euler("x", np.pi)

        return rots



import numpy as np
from scipy.spatial.transform import Rotation as R

from morphosamplers.core import MorphoSampler
from morphosamplers import Sphere
from morphosamplers.sample_types import PoseSet
from morphosamplers.samplers.sphere_samplers.point_sampler import PointSampler


class PoseSampler(MorphoSampler):
    """Sample poses packed on the surface of a sphere with a defined spacing."""
    spacing: float

    def sample(self, obj: Sphere) -> PoseSet:
        # get equally spaced points on surface
        point_sampler = PointSampler(spacing=self.spacing)
        positions = point_sampler.sample(obj)

        # get some initial orientation with z normal to the surface
        z_vectors = positions - np.asarray(obj.center)
        z_vectors /= np.linalg.norm(z_vectors, axis=-1, keepdims=True)
        y_vectors = np.cross(z_vectors, [1, 1, 1])
        y_vectors /= np.linalg.norm(y_vectors, axis=-1, keepdims=True)
        x_vectors = np.cross(y_vectors, z_vectors)
        x_vectors /= np.linalg.norm(x_vectors, axis=-1, keepdims=True)
        orientations = np.empty(shape=(len(positions), 3, 3), dtype=np.float32)
        orientations[:, :, 0] = x_vectors
        orientations[:, :, 1] = y_vectors
        orientations[:, :, 2] = z_vectors

        # randomise in plane rotation
        angles = np.random.uniform(0, 360, size=(len(positions)))
        Rz = R.from_euler('z', angles=angles, degrees=True).as_matrix()
        orientations = orientations @ Rz
        return PoseSet(positions=positions, orientations=orientations)



import napari
import numpy as np

from morphosamplers import Sphere, sphere_samplers

# create a sphere model
sphere = MorphoModels.Sphere(center=(5, 5, 5), radius=20)

# create different types of samplers and use them to sample the sphere
# points
point_sampler = sphere_samplers.PointSampler(spacing=2)
points = point_sampler.sample(sphere)

# poses, z vector normal to the sphere
pose_sampler = sphere_samplers.PoseSampler(spacing=2)
poses = pose_sampler.sample(sphere)


# visualise
viewer = napari.Viewer(ndisplay=3)
viewer.add_points(poses.positions)
viewer.add_vectors(
    data=np.stack(
        [poses.positions, poses.orientations[:, :, 2]],
        axis=1
    ),
    length=5,
    edge_color='orange',
    name='pose z'
)
viewer.add_vectors(
    np.stack(
        [poses.positions, poses.orientations[:, :, 1]],
        axis=1
    ),
    length=2,
    edge_color='cornflowerblue',
    name='pose y',
)
napari.run()


import numpy as np

from morphosamplers.core import MorphoSampler
from morphosamplers import Sphere

from ._sphere_utils import fibonacci_sphere, n_from_spacing


class PointSampler(MorphoSampler):
    spacing: float

    def sample(self, obj: Sphere) -> np.ndarray:
        n = n_from_spacing(self.spacing / obj.radius)
        return fibonacci_sphere(n) * obj.radius + np.asarray(obj.center)





import numpy as np
import einops

GOLDEN_RATIO = (1 + np.sqrt(5)) / 2

# calculate using calculate_best_fit_parameters in dev/sphere_params.py
A = 0.07726122
B = 0.02657319


def fibonacci_sphere(n):
    # http://extremelearning.com.au/how-to-evenly-distribute-points-on-a-
    # sphere-more-effectively-than-the-canonical-fibonacci-lattice/
    i = np.arange(n)
    phi = 2 * np.pi * i / GOLDEN_RATIO
    # this is the version that optimizes for mean distance
    eps = 0.36
    theta = np.arccos(1 - 2 * (i + eps) / (n - 1 + 2 * eps))
    x = np.cos(phi) * np.sin(theta)
    y = np.sin(phi) * np.sin(theta)
    z = np.cos(theta)
    return einops.rearrange([x, y, z], 'xyz b -> b xyz')


def spacing_from_n(n):
    return 1 / np.sqrt(A * n + B)


def n_from_spacing(spacing):
    return int(1 / (spacing ** 2 * A) - B / A)


import numpy as np
import napari

from morphosamplers import Path, path_samplers

# create some control points for a path
xyz = np.linspace([0, 0, 0], [0, 0, 200], num=10)
xyz[:, :2] += np.random.normal(loc=0, scale=10, size=(10, 2))

# make a Path model from these control points
path = Path(control_points=xyz)

### create different types of samplers and use them to sample the path
# points
point_sampler = path_samplers.PointSampler(spacing=10)
positions = point_sampler.sample(path)  # equally spaced positions

# poses
pose_sampler = path_samplers.PoseSampler(spacing=10)
parallel_poses = pose_sampler.sample(path)  # equally spaces parallel poses

# helical poses
helical_pose_sampler = path_samplers.HelicalPoseSampler(spacing=10, twist=30)
helical_poses = helical_pose_sampler.sample(path)  # poses related by 'twist' degrees

### visualise
viewer = napari.Viewer(ndisplay=3)

# positions
viewer.add_points(positions, size=5)

# parallel poses
viewer.add_vectors(
    data=np.stack(
        [parallel_poses.positions, parallel_poses.orientations[:, :, 2]],
        axis=1
    ),
    length=5,
    edge_color='orange',
    name='parallel pose z'
)
viewer.add_vectors(
    np.stack(
        [parallel_poses.positions, parallel_poses.orientations[:, :, 1]],
        axis=1
    ),
    length=2,
    edge_color='cornflowerblue',
    name='parallel pose y',
)

# helical poses
viewer.add_vectors(
    data=np.stack(
        [helical_poses.positions, helical_poses.orientations[:, :, 2]],
        axis=1
    ),
    length=5,
    edge_color='orange',
    name='helical pose z'
)
viewer.add_vectors(
    np.stack(
        [helical_poses.positions, helical_poses.orientations[:, :, 1]],
        axis=1
    ),
    length=2,
    edge_color='cornflowerblue',
    name='helical pose y',
)

# run napari
napari.run()



"""Utility functions."""

import warnings
from typing import Tuple, Union

import numpy as np
from scipy.spatial import KDTree


def _project_vector_onto_plane(vector, plane_normal):
    """Project vector onto a plane defined by its normal.

    Inputs should be normalized; output will be normalized.
    """
    normal_component = np.dot(vector, plane_normal)
    if normal_component == 0:
        # perfectly aligned
        return vector
    proj_plane = vector - normal_component * plane_normal
    proj_plane /= np.linalg.norm(proj_plane)
    return proj_plane


def coaxial_y_vectors_from_z_vectors(
    z: np.ndarray,
    initial_y_vector: Union[np.ndarray, Tuple[float, float, float]] = (
        0.3234,
        0.6543,
        0.978,
    ),
) -> np.ndarray:
    """Calculate y vectors starting from z vectors.

    This function will return the set of unit vectors perpendicular to the z-vectors
    which are maximally coaxial to their neighbours. It assumes that z vectors (n, 3)
    vary smoothly with increasing n.

    Parameters
    ----------
    z : np.ndarray (n, 3)
        The z vectors used to generate y vectors.
    initial_y_vector: Union[np.ndarray, Tuple[float, float, float]]
        The starting y vector projected onto the first z vector to start the projection
        procedure. The default value is weird to make it less likely to match a manual
        entry such as [0, 0, 1] or other simple values.

    Returns
    -------
    y : np.ndarray (n, 3)
        The computed y vectors.
    """
    # normalise z vectors and initialise y
    z = z.copy()
    z /= np.linalg.norm(z, axis=1, keepdims=True)
    y = np.empty((len(z), 3))

    if np.dot(initial_y_vector, z[0]) == 1:
        raise ValueError(
            "cannot generate y vectors because the provided initial_y_vector "
            "and the first z vector are perfectly aligned."
        )
    # normalise initial y vector so that dot product is the projection
    initial_y_vector = initial_y_vector / np.linalg.norm(initial_y_vector)

    # initialise first y
    y[0] = _project_vector_onto_plane(initial_y_vector, z[0])

    # update y vectors in order
    for i in range(len(y) - 1):
        y[i + 1] = _project_vector_onto_plane(y[i], z[i + 1])
    return np.atleast_2d(y)


def deduplicate_points(coords: np.ndarray, exclusion_radius: float) -> np.ndarray:
    """Remove "duplicates" points from an (n, 3) array of coordinates.

    Points are clustered with their neighbours if they are closer than exclusion_radius.
    Clusters are iteratively replaced with their centroid until no clusters remain.
    """
    coords = coords.copy()
    while True:
        tree = KDTree(coords)
        # get clusters sorted by size (biggest first)
        clusters = tree.query_ball_point(coords, exclusion_radius, workers=-1)
        clusters_sorted = sorted(clusters, key=len, reverse=True)
        biggest = clusters_sorted[0]
        if len(biggest) == 1:
            # we reached the degenerate clusters of points with themselves
            break
        centroid = np.mean(coords[biggest], axis=0)
        coords[biggest[0]] = centroid

        mask = np.ones(len(coords), np.bool)
        mask[biggest[1:]] = False
        coords = coords[mask]

    return coords


def minimize_closed_point_row_pair_distance(strips, expected_dist=None):
    # assume closed, circular strips with equal length
    result = [strips[0]]
    ln = len(strips[0])
    for arr, next in zip(strips, strips[1:]):
        best_dist = None
        for i in range(ln):
            next_rolled = np.roll(next, i, axis=0)
            roll_dist = np.linalg.norm(arr - next, axis=1)
            avg_dist = np.mean(roll_dist)
            if best_dist is None or avg_dist < best_dist:
                best_dist = avg_dist
                best_arr = next_rolled
        if expected_dist is not None and best_dist >= expected_dist * np.sqrt(2):
            warnings.warn(
                "The grid is sheared by more than 1 separation in some places",
                stacklevel=2,
            )
        result.append(best_arr)
    return result


def estimate_point_row_directions(rows):
    """
    Guess the directionality of rows of points relative to each other.

    Returns a list of 1 or -1 that can be used as slice steps to index into
    each row and obtain correctly oriented rows.

    Direction is assumed to need inversion if the extrema of two subsequent rows
    are closer to each other when swapped.
    """
    directions = [1]
    for arr, next in zip(rows, rows[1:]):
        start_dist = np.linalg.norm(arr[0] - next[0])
        end_dist = np.linalg.norm(arr[-1] - next[-1])
        dist = start_dist + end_dist
        # now invert direction
        start_dist_inv = np.linalg.norm(arr[0] - next[-1])
        end_dist_inv = np.linalg.norm(arr[-1] - next[0])
        dist_inv = start_dist_inv + end_dist_inv
        if dist_inv < dist:
            # opposite of previous direction
            directions.append(-directions[-1])
        else:
            directions.append(directions[-1])
    return directions


def minimize_point_row_pair_distance(rows, expected_dist=None):
    """Minimize average pair distance at the same index between any number of point rows.

    Given rows of points lying on curves (row, n, xyz), offsets indices in each row in order
    to minimize the euclidean distance between points with the same index in neighbouring rows.

    Once the optimal offsets are found, rows are padded on both ends with nans so they are all the same length.

    For example (- is nan, o is a real point, O are points that participated in the minimum distances):
    -OOOo--
    oOOO---
    -OOOooo
    """
    min_idx = 0
    max_idx = max(len(s) for s in rows)
    offsets = [0]
    # iterate over each row pair
    for arr, next in zip(rows, rows[1:]):
        # pad with enough nans to be able to offset the rows fully without
        # having them roll over to the beginning
        arr_padded = np.pad(
            arr, ((len(next), len(next)), (0, 0)), constant_values=np.nan
        )
        next_padded = np.pad(
            next, ((0, len(next) + len(arr)), (0, 0)), constant_values=np.nan
        )
        tot_len = len(next) + len(arr) + len(next)

        # iterate over all possible offset values,
        best_roll_idx = 0
        best_dist = np.inf
        for i in range(tot_len):
            # roll the next row by 1 relative to the previous iteration
            next_rolled = np.roll(next_padded, i, axis=0)
            # calculate distance for each index pair and average it ignoring nans
            roll_dist = np.linalg.norm(arr_padded - next_rolled, axis=1)
            avg_dist = np.nanmean(roll_dist)

            if np.isnan(avg_dist):
                continue

            if avg_dist < best_dist:
                # if we got a lower average, save this as the best offset
                best_dist = avg_dist
                best_roll_idx = i

        if expected_dist is not None and best_dist >= expected_dist * np.sqrt(2):
            warnings.warn(
                "The grid is sheared by more than 1 separation in some places",
                stacklevel=2,
            )

        # we have the offset that minimises distances
        offset = best_roll_idx - len(next)
        total_offset = offset + offsets[-1]
        offsets.append(total_offset)

    # construct final aligned arrays by offsetting each row by the optimal offset and padding
    # everything so that the lengths are all equal
    aligned = []
    min_idx = min(offsets)
    max_idx = max(o + len(a) for o, a in zip(offsets, rows))
    for arr, offset in zip(rows, offsets):
        padded = np.pad(
            arr,
            ((offset - min_idx, max_idx - offset - len(arr)), (0, 0)),
            mode="constant",
            constant_values=np.nan,
        )
        aligned.append(padded)
    return aligned


def extrapolate_point_rows_with_derivatives(strips, derivatives, separation):
    """
    Extrapolate point strips padded with nans by continuing along their derivatives.

    Extrapolates beyond the first and last finite value in strip continuing
    in the correct direction and spacing by separation.
    """
    extrapolated = []
    for strip, dir in zip(strips, derivatives):
        nans = np.isnan(strip[:, 0])
        left_pad = np.argmax(~nans)
        left_dir = -dir[0]
        left_shift = left_dir / np.linalg.norm(left_dir) * separation
        right_pad = np.argmax(~nans[::-1])
        right_dir = dir[-1]
        right_shift = right_dir / np.linalg.norm(right_dir) * separation
        left_extension = strip[left_pad] + left_shift * np.arange(
            left_pad, 0, -1
        ).reshape(-1, 1)
        right_extension = strip[-right_pad - 1] + right_shift * np.arange(
            1, right_pad + 1
        ).reshape(-1, 1)
        padded = np.concatenate(
            [left_extension, strip[left_pad : -right_pad or None], right_extension]
        )
        extrapolated.append(padded)
    return extrapolated


def within_range(arr, low, high, atol=1e-15):
    """Determine which elements in an array are within a range, with a given tolerance.

    Equivalent to np.where(np.logical_and(arr>=low, arr<=high)), but with tolerance
    for numerical imprecision.
    """
    diff_from_min = arr - low
    diff_from_max = arr - high
    above_min = (diff_from_min >= 0) | np.isclose(diff_from_min, 0, atol=atol)
    below_max = (diff_from_max <= 0) | np.isclose(diff_from_max, 0, atol=atol)
    return above_min & below_max


def strip_true(arr):
    """Strip the leading and trailing False from a bool array."""
    first, last = get_mask_limits(arr)
    return arr[first:last]


def get_mask_limits(mask):
    mask = np.pad(mask, 1)
    beginnings = (mask ^ np.roll(mask, 1)) & mask
    beginnings = np.where(beginnings)[0]
    ends = (mask ^ np.roll(mask, -1)) & mask
    ends = np.where(ends)[0]
    return [(b - 1, e - 1) for b, e in zip(beginnings, ends) if b != e]



"""Tests for NDimensionalSpline."""

import numpy as np
import pytest
from pydantic import ValidationError

from morphosamplers.spline import NDimensionalSpline, Spline3D

n_points = 10
zeros_column = np.zeros((n_points,))
coordinates = np.linspace(0, 1, n_points)
points_2d = np.column_stack([coordinates, coordinates])
expected_points_2d = np.array([[0.5, 0.5], [1, 1]])
points_3d = np.column_stack([zeros_column, coordinates, coordinates])
expected_points_3d = np.array([[0, 0.5, 0.5], [0, 1, 1]])
points_4d = np.column_stack([zeros_column, zeros_column, coordinates, coordinates])
expected_points_4d = np.array([[0, 0, 0.5, 0.5], [0, 0, 1, 1]])


@pytest.mark.parametrize(
    "points,expected_points",
    [
        (points_2d, expected_points_2d),
        (points_3d, expected_points_3d),
        (points_4d, expected_points_4d),
    ],
)
def test_n_dimensional_spline(points, expected_points):
    """Test NDimensionalSpline for 2D, 3D, and 4D splines."""
    spline_model = NDimensionalSpline(points=points, order=4)

    # test that spline order was set
    assert spline_model.order == 4

    # test samping single point
    sample_values = spline_model.sample(0.5)
    expected_single_value = np.atleast_2d(expected_points[0])
    np.testing.assert_allclose(expected_single_value, sample_values)

    # test sampling array of points
    sample_values = spline_model.sample(u=[0.5, 1])
    np.testing.assert_allclose(expected_points, sample_values)


def test_update_spline_points():
    """Test that updating the points recalculates the spline."""
    # line with slope 1
    initial_points = np.array([[0, 0], [0.2, 0.2], [0.3, 0.3], [1, 1]])
    spline_model = NDimensionalSpline(points=initial_points, order=2)

    value_initial_spline = spline_model.sample(u=0.5)
    np.testing.assert_allclose([[0.5, 0.5]], value_initial_spline)

    # line with slope 2
    updated_points = np.array([[0, 0], [0.2, 0.4], [0.3, 0.6], [1, 2]])
    spline_model.points = updated_points
    value_updated_spline = spline_model.sample(u=0.5)
    np.testing.assert_allclose([[0.5, 1]], value_updated_spline)


def test_update_spline_order():
    """Test that updating the order recalculates the spline.

    This does not test for correctness.
    """
    x_coordinate = np.linspace(0, 1, 10)
    y_coordinate = np.power(x_coordinate, 4)
    points = np.column_stack((x_coordinate, y_coordinate))
    initial_spline_order = 2
    spline_model = NDimensionalSpline(points=points, order=initial_spline_order)

    # get the value with the initial spline
    value_initial_spline = spline_model.sample(u=0.2)

    # update the spline order and get value at the same point
    updated_spline_order = 4
    spline_model.order = updated_spline_order
    value_updated_spline = spline_model.sample(u=0.2)

    with pytest.raises(AssertionError):
        # the updated value should be different because the spline is higher order
        np.testing.assert_allclose(value_initial_spline, value_updated_spline)


def test_spline_model_points_list():
    """Spline model should accept a list."""
    points = [
        [0, 1],
        [1, 1],
        [1, 2],
    ]
    _ = NDimensionalSpline(points=points, order=1)


@pytest.mark.parametrize("spline_order", [0.5, -1, 0, 6])
def test_invalid_spline_order(spline_order):
    """Spline order must be an integer in the range [1, 5]."""
    points = np.array([[0, 1], [1, 1], [1, 3], [1, 5], [1, 7], [1, 10], [1, 20]])
    with pytest.raises(ValidationError):
        _ = NDimensionalSpline(points=points, order=spline_order)


equal_points = np.array([[0, 0], [1, 1]])


def test_spline_too_few_points():
    """Number of points should be greater than the spline order."""
    points = np.array([[0, 0], [1, 1]])

    # one point can never work
    with pytest.raises(ValidationError):
        spline = NDimensionalSpline(points=points[:1], order=2)

    # with other numbers of points, the order should be lowered to match
    with pytest.warns(UserWarning):
        spline = NDimensionalSpline(points=points, order=2)
    assert spline.order == len(points) - 1


@pytest.mark.parametrize("derivative_order", [-1, 4])
def test_invalid_spline_derivatives(derivative_order):
    """Derivative order cannot be negative or > spline order."""
    points = np.array([[0, 1], [1, 1], [1, 3], [1, 5], [1, 7], [1, 10], [1, 20]])
    spline_model = NDimensionalSpline(points=points, order=3)

    with pytest.raises(ValueError):
        _ = spline_model.sample(u=0, derivative_order=derivative_order)

    with pytest.raises(ValueError):
        _ = spline_model.sample(
            separation=1, derivative_order=derivative_order
        )


def test_spline_orientations():
    points = np.array([[0, 0, 0], [0, 0, 1], [0, 0, 2], [0, 0, 3]])
    spline = Spline3D(points=points, order=3)
    pt = spline.sample(u=0)
    # lower atol, as it's very strict by default, much more than np.allclose
    np.testing.assert_allclose(pt, 0, atol=1e-8)
    ori = spline.sample_orientations(u=0)
    np.testing.assert_allclose(ori.apply([0, 0, 1]), [[0, 0, 1]])


def test_spline_reverse():
    """After reversing the last input point is the start when sampling."""
    points = np.array([[0, 0, 0], [0, 0, 1], [0, 0, 2], [0, 0, 3]])
    spline = NDimensionalSpline(points=points)
    spline.reverse()
    sample = spline.sample(u=0)
    assert np.allclose(points[-1], sample)



from abc import abstractmethod
from typing import TypeVar, Protocol, Generic

from pydantic.generics import GenericModel

M = TypeVar("M", bound="MorphoModel")
S = TypeVar("S")


class MorphoModel(GenericModel, Generic[M]):
    """A set of attributes defining a geometrical support."""

    class Config:
        allow_mutation = False
        arbitrary_types_allowed = True


class SamplerProtocol(Protocol[M, S]):
    """Protocol for classes that sample a `MorphoModel`."""

    @abstractmethod
    def sample(self, obj: M, *args) -> S:
        """Sample a `MorphoModel` to produces samples of `Sa`."""
        ...


SamplerType = TypeVar("SamplerType", bound=SamplerProtocol)


class MorphoSampler(GenericModel, Generic[SamplerType]):
    """Concrete samplers should subclass this generic model."""

    class Config:
        allow_mutation = False



"""A library for sampling image data along morphological objects."""

from importlib.metadata import PackageNotFoundError, version

try:
    __version__ = version("morphosamplers")
except PackageNotFoundError:
    __version__ = "uninstalled"
__author__ = "Kevin Yamauchi"
__email__ = "kevin.yamauchi@gmail.com"

from morphosamplers.models import Path, Sphere, Surface
from morphosamplers.samplers import path_samplers, sphere_samplers



import numpy as np
from scipy.spatial.transform import Rotation

from morphosamplers.sampler import generate_2d_grid, generate_3d_grid, place_sampling_grids, sample_volume_at_coordinates, sample_volume_along_spline, sample_subvolumes
from morphosamplers.spline import Spline3D


def test_generate_2D_grid():
    shape = (3, 4)
    spacing = (2, 3)
    grid = generate_2d_grid(grid_shape=shape, grid_spacing=spacing)
    assert grid.shape == (*shape, 3)
    # x and y
    for axis in range(2):
        vals = np.swapaxes(grid[..., axis], axis, 0)
        np.testing.assert_allclose(vals[1:] - vals[:-1], spacing[axis])
    # z axis
    z = grid[..., 2]
    assert np.all(z == 0)


def test_generate_3D_grid():
    shape = (2, 5, 3)
    spacing = (2, 1, 3)
    grid = generate_3d_grid(grid_shape=shape, grid_spacing=spacing)
    assert grid.shape == (*shape, 3)
    for axis in range(3):
        vals = np.swapaxes(grid[..., axis], axis, 0)
        np.testing.assert_allclose(vals[1:] - vals[:-1], spacing[axis])


def test_generate_sampling_coordinates():
    grid = generate_2d_grid((2, 2), (1, 1))
    pos = np.zeros((1, 3))
    ori = Rotation.identity(1)
    coords = place_sampling_grids(grid, pos, ori)
    assert coords.shape == (1, 2, 2, 3)


def test_sample_volume_at_coordinates():
    vol = np.zeros((10, 10, 10))
    vol[1, 2, 3] = 1
    coords = np.array([1, 2, 3]).reshape(1, 1, 1, 3)
    sample = sample_volume_at_coordinates(vol, coords)
    np.testing.assert_allclose(sample, 1)


def test_sample_volume_along_spline():
    vol = np.zeros((10, 10, 10))
    vol[1, 2, :] = 1
    pts = np.array([[1, 2, 0], [1, 2, 9]])
    spline = Spline3D(points=pts, order=1)
    sampled = sample_volume_along_spline(vol, spline, sampling_shape=(3, 3), sampling_spacing=1)

    # line of ones
    np.testing.assert_allclose(sampled[:, 1, 1], 1, atol=1e-8)


def test_sample_volume_subvolumes():
    vol = np.zeros((10, 10, 10))
    vol[1, 2, 3] = 1
    pos = np.array([[1, 2, 3]])
    ori = Rotation.identity(1)
    sampled = sample_subvolumes(vol, pos, ori, grid_shape=(3, 3, 3))

    np.testing.assert_allclose(sampled[0, 1, 1, 1], 1, atol=1e-8)



"""Tools for image resampling."""

from typing import Tuple, Union

import numpy as np
import einops
from scipy.ndimage import map_coordinates
from scipy.spatial.transform import Rotation

from .spline import Spline3D
from .surface_spline import GriddedSplineSurface


def generate_3d_grid(
    grid_shape: Tuple[int, int, int] = (10, 10, 10),
    grid_spacing: Tuple[float, float, float] = (1, 1, 1),
) -> np.ndarray:
    """
    Generate a 3D sampling grid with specified shape and spacing.

    The grid generated is centered on the origin, has shape (w, h, d, 3) for
    grid_shape (w, h, d), and spacing grid_spacing between neighboring points.

    Parameters
    ----------
    grid_shape : Tuple[int, int, int]
        The number of grid points along each axis.
    grid_spacing : Tuple[float, float, float]
        Spacing between points in the sampling grid.

    Returns
    -------
    np.ndarray
        Coordinate of points forming the 3D grid.
    """
    # generate a grid of points at each integer from 0 to grid_shape for each dimension
    grid = np.indices(grid_shape).astype(float)
    grid = einops.rearrange(grid, 'xyz w h d -> w h d xyz')
    # shift the grid to be centered on the origin
    grid -= (np.array(grid_shape)) // 2
    # scale the grid to get correct spacing
    grid *= grid_spacing
    return grid


def generate_2d_grid(
    grid_shape: Tuple[int, int] = (10, 10), grid_spacing: Tuple[float, float] = (1, 1)
) -> np.ndarray:
    """
    Generate a 2D sampling grid with specified shape and spacing.

    The grid generated is centered on the origin, lying on the plane with normal
    vector [0, 0, 1], has shape (w, h, 3) for grid_shape (w, h), and spacing
    grid_spacing between neighboring points.

    Parameters
    ----------
    grid_shape : Tuple[int, int]
        The number of grid points along each axis.
    grid_spacing : Tuple[float, float]
        Spacing between points in the sampling grid.

    Returns
    -------
    np.ndarray
        Coordinate of points forming the 2D grid.
    """
    grid = generate_3d_grid(grid_shape=(*grid_shape, 1), grid_spacing=(*grid_spacing, 1))
    return einops.rearrange(grid, 'w h 1 xyz -> w h xyz')


def generate_1d_grid(
    grid_shape: int = 10, grid_spacing: float = 1
) -> np.ndarray:
    """
    Generate a 1D sampling grid with specified shape and spacing.

    The grid generated is centered on the origin, lying along the vector
    [0, 0, 1], has shape (m, 3) for grid_shape m, and spacing grid_spacing
    between neighboring points.

    Parameters
    ----------
    grid_shape : int
        The number of grid points.
    grid_spacing : float
        Spacing between points in the sampling grid.

    Returns
    -------
    np.ndarray
        Coordinate of points forming the 1D grid.
    """
    grid = generate_3d_grid(grid_shape=(1, 1, grid_shape), grid_spacing=(1, 1, grid_spacing))
    return einops.rearrange(grid, '1 1 d xyz -> d xyz')


def place_sampling_grids(
    sampling_grid: np.ndarray, positions: np.ndarray, orientations: Rotation
) -> np.ndarray:
    """
    Copy and transform a given sampling grid onto a set of positions and orientations.

    Returns a (batch, *grid_shape, 3) batch of coordinate grids for sampling.

    Parameters
    ----------
    sampling_grid : np.ndarray
        Grid of points to be copied and transformed.
    positions : np.ndarray
        Array of coordinates used to shift sampling grids.
    orientations : Rotation
        Rotations to be applied to each grid before shifts.

    Returns
    -------
    np.ndarray
        Coordinate of points combining all the transformed grids.
    """
    rotated = []
    grid_shape = sampling_grid.shape
    grid_coords = sampling_grid.reshape(-1, 3)
    # apply each orientation to the grid and store the result
    for orientation in orientations:
        rotated.append(orientation.apply(grid_coords))
    # shift each rotated
    rotated_shifted = np.stack(rotated, axis=1) + positions
    return rotated_shifted.reshape(-1, *grid_shape)


def sample_volume_at_coordinates(
    volume: np.ndarray, coordinates: np.ndarray, interpolation_order: int = 3
) -> np.ndarray:
    """
    Sample a volume with spline interpolation at specific coordinates.

    The output shape is determined by the input coordinate shape such that
    if coordinates have shape (batch, *grid_shape, 3), the output array will have
    shape (*grid_shape, batch).

    Parameters
    ----------
    volume : np.ndarray
        Volume to be sampled.
    coordinates : np.ndarray
        Array of coordinates at which to sample the volume. The shape of this array
        should be (batch, *grid_shape, 3) to allow reshaping back correctly
    interpolation_order : int
        Spline order for image interpolation.

    Returns
    -------
    np.ndarray
        Array of shape (*grid_shape)
    """
    batch, *grid_shape, _ = coordinates.shape
    # map_coordinates wants transposed coordinate array
    sampled_volume = map_coordinates(
        volume, coordinates.reshape(-1, 3).T, order=interpolation_order
    )
    # reshape back (need to invert due to previous transposition)
    sampled_volume = sampled_volume.reshape(*grid_shape, batch)
    # and retranspose to get batch back to the 0th dimension
    return einops.rearrange(sampled_volume, '... batch -> batch ...')


def sample_volume_along_spline(
    volume: np.ndarray,
    spline: Union[np.ndarray, Spline3D],
    sampling_shape: Tuple[int, int] = (10, 10),
    sampling_spacing: float = 1,
    interpolation_order: int = 3,
) -> np.ndarray:
    """
    Extract planes from a volume following a spline path.

    Extracted planes are equidistant along the spline, and perpendicular to
    the spline derivative. The size of the planes is defined by the sampling_shape,
    and the distance between planes and between sample points on the plane is
    equal to sampling_spacing.

    For example: a spline of euclidean length 30, sampling_shape of (10, 10) and spacing of
    2, will result in a sampled volume of shape (15, 10, 10) with twice the pixel size
    of the original volume.

    Parameters
    ----------
    volume : np.ndarray
        Volume to be sampled.
    spline : Spline3D
        Array of points to use to generate the spline, or Spline object,
        along which to sample the volume.
    sampling_shape : Tuple[int, int]
        The number of points along each axis of the grid used for plane sampling.
    sampling_spacing : Tuple[float, float]
        Spacing between points in the sampling grid and along the spline.
    interpolation_order : int
        Spline order for image interpolation.

    Returns
    -------
    np.ndarray
        Sampled volume.
    """
    if not isinstance(spline, Spline3D):
        spline = Spline3D(points=spline)
    positions = spline.sample(separation=sampling_spacing)
    orientations = spline.sample_orientations(separation=sampling_spacing)
    grid = generate_2d_grid(grid_shape=sampling_shape, grid_spacing=(sampling_spacing, sampling_spacing))
    sampling_coords = place_sampling_grids(grid, positions, orientations)
    return sample_volume_at_coordinates(
        volume, sampling_coords, interpolation_order=interpolation_order
    )


def sample_subvolumes(
    volume: np.ndarray,
    positions: np.ndarray,
    orientations: Rotation,
    grid_shape: Tuple[int, int, int] = (10, 10, 10),
    grid_spacing: Tuple[float, float, float] = (1, 1, 1),
    interpolation_order: int = 3,
) -> np.ndarray:
    """
    Extract arbitrarily oriented subvolumes from a volume.

    Parameters
    ----------
    volume : np.ndarray
        Volume to be sampled.
    positions : np.ndarray
        Positions of subvolumes.
    orientations : Rotation
        Orientations of subvolumes.
    grid_shape : Tuple[int, int, int]
        Shape of the 3D grid.
    grid_spacing : Tuple[float, float, float]
        Spacing between points in the sampling grid.
    interpolation_order : int
        Spline order for image interpolation.

    Returns
    -------
    np.ndarray
        Sampled volume.
    """
    grid = generate_3d_grid(grid_shape=grid_shape, grid_spacing=grid_spacing)
    sampling_coords = place_sampling_grids(grid, positions, orientations)
    return sample_volume_at_coordinates(
        volume, sampling_coords, interpolation_order=interpolation_order
    )


def sample_volume_around_surface(
    volume: np.ndarray,
    surface: Union[np.ndarray, GriddedSplineSurface],
    sampling_thickness: int,
    sampling_spacing: float,
    interpolation_order: int = 3,
    masked: bool = False,
) -> np.ndarray:
    """
    Sample a volume around an arbitrary gridded surface.

    For each "root" point on the surface, samples are extracted along a line normal
    to the surface, so that n=sampling_thickness points are extracted, centered on the surface.
    The spacing between points on the surface (unless a pre-generated surface is given) and
    between the sampled points is equal to sampling_spacing. The sampled lines are then
    re-packed into a volume with the same dimensions as the surface, but "flattened".

    For example: a gridded surface of shape (50, 30), sampling_thickness of 10 and spacing of 3,
    will result in a sampled volume of shape (50, 30, 10) with three times the pixel size
    of the original volume.

    Parameters
    ----------
    volume : np.ndarray
        Volume to be sampled.
    surface : Union[np.ndarray, SplineSurfaceGrid]
        Array of points to be used to generate a surface, or SplineSurfaceGrid object,
        along which to sample the volume.
    sampling_thickness : int
        Thickness of sampled surface in pixels.
    sampling_spacing : float
        Spacing between sampled pixels.
    interpolation_order : int
        Spline order for image interpolation.

    Returns
    -------
    np.ndarray
        Sampled volume.
    """
    if not isinstance(surface, GriddedSplineSurface):
        surface = GriddedSplineSurface(points=surface, separation=sampling_spacing)
    positions = surface.sample()
    orientations = surface.sample_orientations()
    grid = generate_1d_grid(grid_shape=sampling_thickness, grid_spacing=sampling_spacing)
    sampling_coords = place_sampling_grids(grid, positions, orientations)
    sampled = sample_volume_at_coordinates(volume, sampling_coords, interpolation_order=interpolation_order)
    if masked:
        sampled[~surface.mask] = np.nan
    return sampled.reshape(*surface.grid_shape, sampling_thickness)



from typing import Tuple

import einops
import numpy as np
from scipy.interpolate import splprep, splev

from morphosamplers import Path
from morphosamplers.core import MorphoSampler
from morphosamplers.sample_types import Points


class PointSampler(MorphoSampler):
    spacing: float

    def sample(self, obj: Path) -> Points:
        """Sample a `Path` to produces an `(n, 3)` array of points."""
        tck, total_length = self.prepare_spline(obj)
        n_samples = total_length // self.spacing
        max_u = 1 - ((total_length % self.spacing) / total_length)
        u = np.linspace(0, max_u, num=int(n_samples))
        return einops.rearrange(splev(u, tck), 'xyz b -> b xyz')

    @staticmethod
    def prepare_spline(
        path: Path, n_initial_samples: int = 10_000
    ) -> Tuple[Tuple[np.ndarray], float]:
        # oversample an initial spline between control points
        points = einops.rearrange(path.control_points, 'b xyz -> xyz b')
        u = np.linspace(0, 1, num=n_initial_samples)
        spline_order = 3 if len(path) > 3 else len(path) - 1
        tck, _ = splprep(points, s=0, k=spline_order)
        samples = einops.rearrange(splev(u, tck), 'xyz b -> b xyz')

        # calculate the cumulative length as we move along the path.
        inter_point_differences = np.diff(samples, axis=0)
        inter_point_distances = np.linalg.norm(inter_point_differences, axis=1)
        cumulative_distance = np.cumsum(inter_point_distances)
        total_length = float(cumulative_distance[-1])
        cumulative_distance /= total_length  # normalise to unit length
        cumulative_distance = np.insert(cumulative_distance, 0, 0)

        # equidistant samples in u yield equidistant samples in euclidean space
        tck, _ = splprep(samples.T, u=cumulative_distance, s=0, k=spline_order)
        return tck, total_length



from typing import Tuple

from pydantic import validator

from morphosamplers.core import MorphoModel


class Sphere(MorphoModel):
    """A sphere in 3D."""
    center: Tuple[float, float, float]
    radius: float

    @validator('center', pre=True)
    def coerce_to_tuple(cls, value):
        return tuple(value)

